#!/usr/bin/python
# -*- coding: utf-8 -*-

import json
import inspect

###################################################################################################
# supy.serialize.py
#
# Simple encoder/decoder classes for serializing python objects and python classes 
#
# Encoding works for objects of arbitrary classes
#
# However, creation of objects during decoding requires some conventions:
# - Decoder creates an object by calling the creator (__init__) with arguments corresponding
#   to all object fields
# - Thus, the creators (__init__ method) must offer parameters with names being identical to 
#   the relevant object fields
# - if classes have attribute fields set inside the constructor 
#   (that is, the fields are not given as parameters to __init__) 
#   then the constructor __init__ must have an **args parameters 
#   in order to catch additional parameters that are not used
# For further reference see the examples in the module test below...
# 
# Code inspired from a tutorial on http://pymotw.com/2/json/
# 
# 28.3.2013 - 7.4.2013 by Andreas Knoblauch
###################################################################################################


###################################################################################################
# SupyEncoder: Encoder class
###################################################################################################
class SupyEncoder(json.JSONEncoder):
    
    def default(self, obj):
        if inspect.isclass(obj):
            d = {'__isclass__':True, 
                 '__class__':obj.__name__, 
                 '__module__':obj.__module__,
                 }
        else:
            d = {'__isclass__':False, 
                 '__class__':obj.__class__.__name__, 
                 '__module__':obj.__module__,
                 }
            d.update(obj.__dict__)
        if d['__class__']=="instancemethod":     # so far do not consider methods
            return None
        else:
            return d


###################################################################################################
# SupyDecoder: Decoder class
###################################################################################################
class SupyDecoder(json.JSONDecoder):
    
    def __init__(self,**args):
        json.JSONDecoder.__init__(self,object_hook=self.dict_to_object,**args)

    def dict_to_object(self, d):
        if '__class__' in d:
            class_name = d.pop('__class__')
            module_name = d.pop('__module__')
            module = __import__(module_name)
            class_ = getattr(module, class_name)
            if d.pop('__isclass__'):
                obj = class_
            else:
                args = dict( (key.encode('ascii'), value) for key, value in d.items())
                obj = class_(**args)
        else:
            obj = d
        return obj


#######################################
# Module test     
#######################################
if __name__ == '__main__':

    print "#######################################"
    print "# Test for module supy.serialize.py    "
    print "#######################################"

    class obj_cfg:
        a=5
        b=7
        c=10

    class MyObj(object):                  # simple class, all object fields given in __init__
        def __init__(self, s):
            self.s = s
        def __repr__(self):
            return '<MyObj(%s)>' % self.s

    class MyObj2(object):                 # ditto, but cfg may be a class parameter
        def __init__(self, s, cfg=None):
            self.s = s
            self.cfg=cfg
        def __repr__(self):
            return '<MyObj(%s)>' % self.s

    class MyObj3(object):                 # as MyObj2, but __init__ defines additional fields not given in the parameter list 
        def __init__(self, s, cfg=None, **args):      # **args is required because __init__ sets additional fields 
            self.s = s
            self.cfg=cfg
            self.f1 = 245
            self.f2 = 456

        def __repr__(self):
            return '<MyObj(%s)>' % self.s

    print "\n\n##### (i) encoding/decoding directly with the supy classes #####"
    obj = MyObj('internal data')
    obj2 = MyObj2(obj,cfg=obj_cfg)
    obj3 = MyObj3(obj,cfg=obj_cfg)
    print "obj=",obj
    print "obj2=",obj2
    print "obj3=",obj3
    str_obj = SupyEncoder().encode(obj)
    str_obj2 = SupyEncoder().encode(obj2)
    str_obj3 = SupyEncoder().encode(obj3)
    print "encoded obj=", str_obj
    print "encoded obj2=", str_obj2
    print "encoded obj3=", str_obj3

    obj_ = SupyDecoder().decode(str_obj)
    print "decoded obj_=",obj_
    obj2_ = SupyDecoder().decode(str_obj2)
    print "decoded obj2_=",obj2_
    obj3_ = SupyDecoder().decode(str_obj3)
    print "decoded obj3_=",obj3_

    print "\n\n##### (ii) encoding/decoding with json.dumps/loads #####"
    obj = MyObj('internal data')
    obj2 = MyObj2(obj,cfg=obj_cfg)
    obj3 = MyObj3(obj2,cfg=obj_cfg)
    obj4 = [1,2,3,4]
    print "obj3=",obj3
    str_obj3 = json.dumps(obj3,cls=SupyEncoder)
    print "encoded obj3=", str_obj3
    obj3_ = json.loads(str_obj3,cls=SupyDecoder)
    print "decoded obj3_=",obj3_
    print "\nobj4=",obj4
    str_obj4 = json.dumps(obj4,cls=SupyEncoder)
    print "encoded obj4=", str_obj4
    obj4_ = json.loads(str_obj4,cls=SupyDecoder)
    print "decoded obj4_=",obj4_
    



#!/usr/bin/python
# -*- coding: utf-8 -*-

from supy.sqlforms import *



##################################################################################################################################################################
##################################################################################################################################################################
# Part I: SQLTableEditor 
#         classes: SQLTableEditor, SQLTableEditor_config
##################################################################################################################################################################
##################################################################################################################################################################

################################################################################################
# SQLTableEditor_config
# provides parameters for SQLTableEditor 
################################################################################################
class SQLTableEditor_config:
    table_cfg            = None               # table to be edited, first column is assumed to be primary key of type INTEGER!!!
    pkeys_readonly       = []                 # records with primary key in this list are considered as READ_ONLY (e.g., for default records)
    min_max_records      = None               # or [nMin, nMax]; if !=None this specifies the minimum and maximum number of data records in the table
    cfg_choiceLB         = None               # configuration for choice listbox (typically a SQLListbox_config)
    cfg_recForm          = None               # configuration for record form (typically a SupySQLForm_config)
    align_widgets        = 'horizontal'       # either 'vertical' or 'horizontal' for alignment of listbox and form
    lbButtons_str        = ['New', 'Copy', 'Delete']            # text for listbox buttons
    recFormButtons_str   = ['Reset', 'Default', 'Commit']       # text for form buttons
    ctrlRadioButtons_str = ['View', 'Edit']                     # text for radiobuttons (read-only, write/edit option)
    ctrlButtons_str      = ['Commit & Close', 'Cancel & Close'] # text for control buttons
    text_showinfo_cannot_delete_readonly = ("Info", "Cannot delete read-only data record!")
    text_askyesno_delete = ("Delete data record?", "Do you really want to delete this data record? \nData record : ")
    # extend tbed_.... to get a valid initialization of a new record that is conditioned on exam.key (because examofcourse has no direct link to exam)
    ntom_init_on_new     = None    # list of [col_to_be_initialized,col_initialized_by,where_cond]

###################################################################################################################
# SQLTableEditor
# consists of:
#    (1) choiceLB: a listbox to choose one data record from a database table (e.g., a SQLListbox)
#                  !!! first data entry of listbox should correspond to primary key of the table to be edited !!!
#    (2) recForm: a form to edit the record (e.g., a SupySQLForm)
#    (3) lbButtons: buttons to control the database table via the listbox:
#                   - New: Insert a new (default) record into the table
#                   - Copy: Insert a copy of the selected data recrod into the table
#                   - Delete: Delete the selected data record
#   (4) formButtons: buttons to control the record form:
#                   - Reset/Rollback: reset data record to the values of the last commit
#                   - Default: Set data record to default values
#                   - Commit: Commit data record to current input values
#   (5) ctrlRadioButtons: radio buttons to control the SQLTableEditor
#                    - view: read-only mode 
#                    - edit: edit (write) mode 
#   (6) ctrlButtons: buttons to control the SQLTableEditor
#                    - Commit & Close: Commit last inputs and close editor
#                    - Cancel & Close: Cancel last inputs and close editor
#   Extension as of 19/9/2016:
#   (7) extrabuttons = [[buttontext, buttonlocation, action],...]
#         - buttontext is text on buttons
#         - buttonlocation is currently only 'listboxleft' (on the left side of the listbox) possible
#         - action is a method that will be called if the button is pressed
###################################################################################################################

class SQLTableEditorFrame(Frame): 
    def __init__(self, parent, sqldatabase, cfg=SQLTableEditor_config, cond_on=[], extrabuttons=[]):
        Frame.__init__(self,parent)  
        # (i) basic initialization and flags
        self.parent=parent
        self.sqldatabase = sqldatabase
        self.cfg         = cfg
        self.conditioned_on = cond_on
        self.flag_readonly           =1  # default is read-only model (also set by self.on_ctrl_mode_readonly()
        self.extrabuttons_listboxleft=[e for e in extrabuttons if e[1]=='listboxleft']
        # (ii) allocate widgets
        # (ii.a) record form
        self.recFormFrame = Frame(self)
        self.recForm=SupySQLFormFrame(self.recFormFrame,self.sqldatabase,self.cfg.cfg_recForm,conditioned_on=cond_on)
        self.rfButtons_frame  = Frame(self.recFormFrame)
        self.rfButton_Reset   = Button(self.rfButtons_frame,text=self.cfg.recFormButtons_str[0],command=self.on_rf_reset)
        self.rfButton_Default = Button(self.rfButtons_frame,text=self.cfg.recFormButtons_str[1],command=self.on_rf_default)
        self.rfButton_Commit  = Button(self.rfButtons_frame,text=self.cfg.recFormButtons_str[2],command=self.on_rf_commit)
        # (ii.b) listbox 
        self.lbFrame=Frame(self)
        self.choiceLB=SQLListbox(self.lbFrame,self.sqldatabase,self.cfg.cfg_choiceLB,callback_select=self.on_lb_select,conditioned_on=cond_on)
        self.lbButtons_frame  = Frame(self)
        self.lbButtons_extra_left = [Button(self.lbButtons_frame,text=e[0],command=lambda : self.onExtraButton(e)) for e in self.extrabuttons_listboxleft]
        self.lbButtons_New    = Button(self.lbButtons_frame,text=self.cfg.lbButtons_str[0],command=self.on_lb_new)
        self.lbButtons_Copy   = Button(self.lbButtons_frame,text=self.cfg.lbButtons_str[1],command=self.on_lb_copy)
        self.lbButtons_Delete = Button(self.lbButtons_frame,text=self.cfg.lbButtons_str[2],command=self.on_lb_delete)
        # (ii.c) control panel
        self.ctrl_frame = Frame(self)
        self.var_ctrlRadioButton = IntVar()
        self.var_ctrlRadioButton.set(0)
        self.ctrlRadioButton_frame = Frame(self.ctrl_frame)
        self.ctrlRadioButton_view = Radiobutton(self.ctrlRadioButton_frame, text=self.cfg.ctrlRadioButtons_str[0], command=self.on_ctrl_mode_readonly,variable=self.var_ctrlRadioButton,value=0)
        self.ctrlRadioButton_edit = Radiobutton(self.ctrlRadioButton_frame, text=self.cfg.ctrlRadioButtons_str[1], command=self.on_ctrl_mode_edit    ,variable=self.var_ctrlRadioButton,value=1)
        self.ctrlButtons_frame = Frame(self.ctrl_frame)
        self.ctrlButton_CommitClose = Button(self.ctrlButtons_frame,text=self.cfg.ctrlButtons_str[0], command=self.on_ctrl_commitclose)
        self.ctrlButton_CancelClose = Button(self.ctrlButtons_frame,text=self.cfg.ctrlButtons_str[1], command=self.on_ctrl_cancelclose)
        # (ii.d) packing
        if self.cfg.align_widgets=='horizontal':
            # align widgets horizontally
            self.lbFrame.grid(row=0,column=0)          # lbFrame on self
            self.choiceLB.pack(side=TOP)               # lb on lbframe
            self.lbButtons_frame.grid(row=1, column=0) # lbButtons on self
            for b in self.lbButtons_extra_left: b.pack(side=LEFT)   # extrabuttons?
            self.lbButtons_New.pack(side=LEFT)         # buttons on lbButtons
            self.lbButtons_Copy.pack(side=LEFT)
            self.lbButtons_Delete.pack(side=LEFT)
            self.recFormFrame.grid(row=0,column=1)     # recFormFrame on self
            self.recForm.pack(side=TOP)
            self.rfButtons_frame.pack(side=BOTTOM)
            self.rfButton_Reset.pack(side=LEFT)
            self.rfButton_Default.pack(side=LEFT)
            self.rfButton_Commit.pack(side=LEFT)
            self.ctrl_frame.grid(row=1, column=1)      # ctrl panel on self
            self.ctrlButtons_frame.pack(side=RIGHT)    # ctrlButtonsFrame on ctrl panel
            self.ctrlButton_CancelClose.pack(side=RIGHT)
            self.ctrlButton_CommitClose.pack(side=RIGHT)
            self.ctrlRadioButton_frame.pack(side=RIGHT) # radiobuttons on ctrl panel 
            self.ctrlRadioButton_view.pack(side=LEFT)
            self.ctrlRadioButton_edit.pack(side=LEFT)
        else:
            # align widgets vertically
            pass
        # (iii) finish initialization of transaction control
        self.on_ctrl_mode_readonly()    # default is read-only mode
        
    """
    get_new_primary_key 
     - get new primary key for inserting a new data record into table
     - assumes that the first column of the table is the primary key!!!
     - here: assumes that the primary key has INTEGER-type
    """
    def get_new_primary_key(self):
        table_name = self.cfg.table_cfg.name          # name of the table
        pkey_col   = self.cfg.table_cfg.col_names[0]  # column name of primary key
        return self.sqldatabase.get_new_primary_key(table_name,pkey_col)[0]

    """
    commit_inputs
     - is called when inputs are committed
     - checks if there is an active transaction and commmits inputs (and sets transaction flag to zero)
    """
    def commit_inputs(self):
        if self.flag_readonly==0:  
            # (i) commit data of input widgets to database
            self.recForm.commit_inputs()
            # (ii) update listbox widget 
            table_name = self.cfg.table_cfg.name          # name of the table
            pkey_col   = self.cfg.table_cfg.col_names[0]  # column name of primary key
            where_expr = None                             # just in case of an empty table...
            if(self.choiceLB.selected_data!=None)and(self.choiceLB.selected_data!=[])and(len(self.choiceLB.selected_data[0])>0):
                pk = self.choiceLB.selected_data[0][0]        # it is assumed that primary key corresponds to first entry of listbox data
                where_expr = table_name+"."+pkey_col + " = " + str(pk)
            self.choiceLB.setListboxKeys(where_expr,flag_callback=False, flag_ignoreoldselection=True)   # update listbox, but no callback (to avoid recursion loop)
            
        
    """
    on_lb_select
     - is called if a selection from the choice listbox has been made
     - then: - save data from recFormFrame to database
             - display selected data
    """
    def on_lb_select(self,selected_data_idx,selected_data,selected_data_str):
        # (i) commit old input data (if necessary)
        self.commit_inputs()       # commit old transaction (if there is any)
        # (ii) get new selection and begin transaction (if necessary)
        if(selected_data!=None)and(selected_data!=[])and(len(selected_data[0])>0):
            self.recForm.setData([self.cfg.cfg_choiceLB.cols[0]],[selected_data[0][0]])

    """
    on_lb_new
     - is called on pressing the "New" Button 
     - then: - insert a new record into database (with new primary key and default values)
             - select new data record (and save old data from recForm)
    """
    def on_lb_new(self):
        if (self.cfg.min_max_records!=None) and (len(self.cfg.min_max_records)==2): 
            if self.choiceLB.data!=None and len(self.choiceLB.data)>=self.cfg.min_max_records[1]: return;     # ignore new-command if there are already the maximum number of records!
        success=0    # flag for success to commit new record
        # (i) commit old input data (if necessary)
        self.commit_inputs()       # commit old transaction (if there is any)
        # (ii) insert default record
        table_name = self.cfg.table_cfg.name          # name of the table
        pkey_col   = self.cfg.table_cfg.col_names[0]  # column name of primary key
        pk = self.get_new_primary_key()               # new primary key
        #print(" new primary key pk=", pk)
        cmd = 'insert into ' + table_name + ' (' + pkey_col + ' ) values (' + str(pk) + ')' 
        self.sqldatabase.begin_transaction()
        try:
            if(self.sqldatabase.debugLevel>0): print("EXECUTE SQL COMMAND: ", cmd)
            self.sqldatabase.curs.execute(cmd)
        except:
            str_error = "Insert Exception while inserting new record in table " + str(table_name) + ": \n\n" + str(sys.exc_info()) + \
                        "\n\nInsert query will be rolled-back."
            tkinter.messagebox.showerror(title="Insert Error", parent=self, message=str_error)
            self.sqldatabase.rollback_transaction()
        else:
            # (iii) set inherited (preconditioned) values
            print("\ncond.on=",self.conditioned_on)
            print("\nchoiceLB.tables=",self.choiceLB.tables)
            print("\nchoicLB.join_on=",self.choiceLB.join_on)
            c1,v1=getUniqueListOfPreconditions([],[],self.conditioned_on,self.choiceLB.tables,self.choiceLB.join_on)
            #c,v=[i[0] for i in self.conditioned_on], [i[1] for i in self.conditioned_on]
            print("\n\nafter getUniqueListOfPreconditions: onlb_new1: c1,v1=",c1,v1)
            c,v=getPreconditions4tables(c1,v1,[table_name])
            print("\n\nonlb_new2: c,v=",c,v)
            if len(c)>0: self.sqldatabase.simple_update_byPKEY(table_name,c,v,[pkey_col],[str(pk)])
            # (iv) check if there are (additional) ntom initializations
            if self.cfg.ntom_init_on_new==None:   
                success=1
            else:
                for jj in range(len(self.cfg.ntom_init_on_new)):
                    inon=self.cfg.ntom_init_on_new[jj]
                    assert isinstance(inon,list) and (len(inon)==4) and (not "." in inon[0]) and ("." in inon[1]) and ("." in inon[2]) and ("." in inon[3]), "Wrong format of list ntom_init_on_new=" + str(self.cfg.ntom_init_on_new) + \
                            " !\nInstead, each list entry self.cfg.ntom_init_on_new[i] must have format [col_in_tab2_to_be_initialized, tabnm.key, tabnm.key_to_tab1, tab1.key]" 
                    col_in_tab2_to_be_initialized=inon[0]
                    tabnm_key = inon[1]
                    tabnm_key_to_tab1 = inon[2]
                    tab1_key = inon[3]
                    tabnm=tabnm_key.split(".")[0]
                    keynm=tabnm_key.split(".")[1]
                    assert tab1_key in c1, "Cannot find tab1.key="+str(tab1_key)+" in preconditions " + str(c1) + " for ntom_init_on_new=" + str(self.cfg.ntom_init_on_new) + " !" 
                    val1_key = [v1[i] for i in range(len(c1)) if c1[i]==tab1_key][0]
                    wc = self.sqldatabase.getWhereClause_from_ColumnValues([tabnm],[tabnm_key_to_tab1],[val1_key])
                    res = self.sqldatabase.simple_select([tabnm_key],[tabnm],where=wc)
                    if (res==None) or (len(res)==0):
                        str_error="Cannot insert new record because there is no appropriate initialization!\n" + self.cfg.ntom_init_on_new_errormsg
                        tkinter.messagebox.showerror(title="Insert Error", parent=self, message=self.cfg.ntom_init_on_new_errormsg[jj])
                        self.sqldatabase.rollback_transaction()
                    else:
                        self.sqldatabase.simple_update_byPKEY(table_name,[col_in_tab2_to_be_initialized],[res[0][0]],[pkey_col],[str(pk)])
                        success=1
        # (v) select new data record
        if(success>0):
            self.sqldatabase.commit_transaction()
            self.choiceLB.setListboxKeys(table_name+"."+pkey_col + " = " + str(pk), flag_ignoreoldselection=True)

    """
    on_lb_copy
     - is called on pressing the "Copy" Button 
     - then: - insert a new record into database being a copy of the current data record 
             - select new data record (and save old data from recForm)
    """
    def on_lb_copy(self):
        if (self.cfg.min_max_records!=None) and (len(self.cfg.min_max_records)==2): 
            if self.choiceLB.data!=None and len(self.choiceLB.data)>=self.cfg.min_max_records[1]: return;     # ignore copy-command if there are already the maximum number of records!
        if (self.choiceLB.selected_data!=None) and (len(self.choiceLB.selected_data)>0) and (len(self.choiceLB.selected_data[0])>0): # is there selected data to be copied??
            # (i) commit old input data (if necessary)
            self.commit_inputs()       # commit old transaction (if there is any)
            # (ii) insert new record that contains copy of the current data
            table_name = self.cfg.table_cfg.name          # name of the table
            table_cols = self.cfg.table_cfg.col_names     # columns of the table
            pkey_col   = self.cfg.table_cfg.col_names[0]  # column name of primary key
            pk_new = self.get_new_primary_key()           # new primary key
            pk_old = self.choiceLB.selected_data[0][0]    # it is assumed that primary key corresponds to first entry of listbox data
            cols2copy = [c for c in table_cols[1:] if (self.cfg.cfg_recForm.cols2copy==None) or (c in self.cfg.cfg_recForm.cols2copy)]  # columns to be copied (without primarky key)
            cols2copy = [pkey_col] + cols2copy    # columns to be copied (with primary key at pos 0)
            wc=self.sqldatabase.getWhereClause_from_ColumnValues([table_name],[pkey_col],[pk_old])
            rec=self.sqldatabase.simple_select(cols2copy,[table_name],where=wc)      # old column data (including primary key at pos 0)
            assert (len(rec)==1)and(len(rec[0])==len(cols2copy)), \
                " Cannot create copy of columns " + cols2copy + " for table " + str(table_name) + " because of primary key mismatch!\nResult record for primary key=" + str(pk_old) + " is " + str(rec)
            rec=list(rec[0])                              # copy of record
            rec[0]=pk_new                                 # replace primary key by new one
            self.sqldatabase.begin_transaction()          # do insertion transaction
            try:
                self.sqldatabase.simple_insert(table_name,cols2copy,rec)    # insert new record
            except:
                str_error = "Insert Exception while inserting copy record in table " + str(table_name) + ": \n\n" + str(sys.exc_info()) + \
                            "\n\nInsert query will be rolled-back."
                tkinter.messagebox.showerror(title="Insert Error", parent=self, message=str_error)
                self.sqldatabase.rollback_transaction()
            else:
                self.sqldatabase.commit_transaction()
            # (iii) copy cascade in related tables (by ref_ntom)
            form_cfg=self.cfg.cfg_recForm
            for i in range(len(form_cfg.cols)):
                if (form_cfg.cols[i]==None)and(form_cfg.cols_type[i]=='ref_ntom')and(form_cfg.cols_readonly[i]==0):
                    ntom_cc=form_cfg.cols_ref[i].ntom_copy_cascade    # defnition of copy cascade behavior in format:
                    if (ntom_cc!=None):                               #     [(table1,key1),(tablenm,keynm,foreignkey,[list of columns to be copied])]
                        # (iii.1) decode ntom_copy_cascade parameters...
                        cc_tab1 ,cc_key1                                    = ntom_cc[0][0],ntom_cc[0][1]
                        cc_tabnm,cc_keynm,cc_keynm_foreigntab1,cc_cols2copy = ntom_cc[1][0],ntom_cc[1][1],ntom_cc[1][2],ntom_cc[1][3]
                        assert cc_tab1==table_name and cc_key1==pkey_col, "Mismatch during copy cascade of table "+table_name+" with primary key " + pkey_col + " with invalid ntom_copy_cascade=" + str(ntom_cc) + "!"
                        # (iii.2) get all records of tabnm that are associated with old record in tab1 where key1=pk_old
                        wc=self.sqldatabase.getWhereClause_from_ColumnValues([cc_tabnm],[cc_keynm_foreigntab1],[pk_old])
                        oldrecs=self.sqldatabase.simple_select(cc_cols2copy,[cc_tabnm],where=wc)      # records in tabnm associated with tab1.pk_old
                        print("\n\n oldrecs=", str(oldrecs))
                        # (iii.3) for each such record, insert a copy associated with tab1.pk_new
                        for orc in oldrecs:
                            print("\norc = ", orc)
                            cc_tabnm_pk_new = self.sqldatabase.get_new_primary_key(cc_tabnm,cc_keynm)[0]    # new primary key for tabnm
                            self.sqldatabase.simple_insert(cc_tabnm, [cc_keynm,cc_keynm_foreigntab1]+cc_cols2copy, [cc_tabnm_pk_new,pk_new]+list(orc))
            # (iv) select new data record
            self.choiceLB.setListboxKeys(table_name+"."+pkey_col + " = " + str(pk_new), flag_ignoreoldselection=True)

    """
    on_lb_delete
     - delete selected data record 
     - then: - select next data record (of listbox)
    """
    def on_lb_delete(self):
        if (self.cfg.min_max_records!=None) and (len(self.cfg.min_max_records)==2): 
            if self.choiceLB.data!=None and len(self.choiceLB.data)<=self.cfg.min_max_records[0]: return;     # ignore delete-command if there are only the minimum number of records in the table!
        if (self.choiceLB.selected_data!=None) and (len(self.choiceLB.selected_data)>0) and (len(self.choiceLB.selected_data[0])>0): # is there selected data to be deleted??
            # (i) check if data record should really be deleted
            pk = self.choiceLB.selected_data[0][0]       # it is assumed that primary key corresponds to first entry of listbox data
            if(self.cfg.pkeys_readonly!=None)and(pk in self.cfg.pkeys_readonly):
                tkinter.messagebox.showinfo(self.cfg.text_showinfo_cannot_delete_readonly[0], self.cfg.text_showinfo_cannot_delete_readonly[1], parent=self)
            elif tkinter.messagebox.askyesno(self.cfg.text_askyesno_delete[0],self.cfg.text_askyesno_delete[1] + self.choiceLB.selected_data_str[0] + " ?"):
                # (ii) delete data record
                table_name = self.cfg.table_cfg.name          # name of the table
                pkey_col   = self.cfg.table_cfg.col_names[0]  # column name of primary key
                self.sqldatabase.begin_transaction()          # do delete transaction
                try:
                    self.sqldatabase.simple_delete(table_name,[pkey_col],[pk])
                except:
                    str_error = "Delete Exception while deleting record in table " + str(table_name) + ": \n\n" + str(sys.exc_info()) + \
                                "\n\nDelete query will be rolled-back."
                    tkinter.messagebox.showerror(title="Delete Error", parent=self, message=str_error)
                    self.sqldatabase.rollback_transaction()
                else:
                    self.sqldatabase.commit_transaction()
                # (iii) select neighboring data record
                i=self.choiceLB.selected_data_idx[0]
                pk_next=None
                if len(self.choiceLB.data)>(i+1):
                    pk_next = self.choiceLB.data[i+1][0]
                elif i>0:
                    pk_next = self.choiceLB.data[i-1][0]
                wc=None
                if(pk_next!=None): wc = table_name+"."+pkey_col + " = " + str(pk_next)
                self.choiceLB.setListboxKeys(wc, flag_ignoreoldselection=True)

    """
    on_rf_reset
     - reset (non-primary-key) form data to values stored in database 
    """
    def on_rf_reset(self):
        if(self.choiceLB.selected_data!=None) and (len(self.choiceLB.selected_data)>0):
            self.recForm.setData([self.cfg.cfg_choiceLB.cols[0]],[self.choiceLB.selected_data[0][0]])

    """
    on_rf_default
     - reset (non-primary-key) form data to default values  
    """
    def on_rf_default(self):
        rf_cfg=self.cfg.cfg_recForm
        data_default = self.sqldatabase.simple_select_defaultvalues(rf_cfg.cols,rf_cfg.tables)
        print("data_default=", data_default)
        for i in range(len(data_default)):
            if (i>0) and (data_default[i]!=None):                # skip first (primary key) column and None-data
                self.recForm.setEntryData(i,data_default[i])

    """
    on_rf_commit
     - commit form data data into sql database   
    """
    def on_rf_commit(self):
        self.commit_inputs()

    """
    on_ctrl_mode_readonly
     - enter read-only mode    
    """
    def on_ctrl_mode_readonly(self):
        # (i) commit old inputs (if necessary) and set readonly flag
        self.commit_inputs()
        self.flag_readonly=1
        # (ii) disable record form widgets
        self.recForm.disableWidgets()
        self.rfButton_Reset.configure(state=DISABLED)
        self.rfButton_Default.configure(state=DISABLED)
        self.rfButton_Commit.configure(state=DISABLED)
        # (iii) disable listbox widgets
        self.lbButtons_New.configure(state=DISABLED)
        self.lbButtons_Copy.configure(state=DISABLED)
        self.lbButtons_Delete.configure(state=DISABLED)
        # (iv) disable control panel widgets
        self.ctrlButton_CommitClose.configure(state=DISABLED)

    """
    on_ctrl_mode_edit
     - enter edit (read&write) mode    
    """
    def on_ctrl_mode_edit(self):
        # (i) set readonly flag
        self.flag_readonly=0
        # (ii) disable record form widgets
        self.recForm.enableWidgets()
        self.rfButton_Reset.configure(state=NORMAL)
        self.rfButton_Default.configure(state=NORMAL)
        self.rfButton_Commit.configure(state=NORMAL)
        # (iii) disable listbox widgets
        self.lbButtons_New.configure(state=NORMAL)
        self.lbButtons_Copy.configure(state=NORMAL)
        self.lbButtons_Delete.configure(state=NORMAL)
        # (iv) disable control panel widgets
        self.ctrlButton_CommitClose.configure(state=NORMAL)

    """
    on_ctrl_commitclose
     - commit form values and close     
    """
    def on_ctrl_commitclose(self):
        self.recForm.commit_inputs()
        self.parent.commitclose=1
        self.destroy()

    """
    on_ctrl_cancelclose
     - cancel form values and close     
    """
    def on_ctrl_cancelclose(self):
        self.destroy()

    """
    onExtraButton
     - extra actions if one of the extra buttons have been pressed
     - e is in the format [buttontext, buttonlocation, action]
     - buttontext is actually ignored!     
    """
    def onExtraButton(self,e):
        assert isinstance(e,(list,tuple)) and len(e)==3, "extra-action description must be in format [buttontext, buttonlocation, action]"
        assert e[1] in ['listboxleft'], "Unknown button location "+str(e[1])
        f=e[2]
        if e[1]=='listboxleft': f(self.choiceLB) 

def editSQLTables(parent, sqldatabase, title=None, cfg=SQLTableEditor_config,cond_on=[], extrabuttons=[], flagIndepFromParent=0):  # for parameters see SQLListboxForm_config
    if title==None: title=cfg.title
    top=Toplevel()
    top.commitclose=0
    if(flagIndepFromParent<=0): 
        top.grab_set()    # make window modal (i.e., redirect all inputs from parent window to listbox dialog)
        if parent: top.transient(parent)       # make window transient, e.g., minmize with parent etc.
    if title: top.title(title) 
    f=SQLTableEditorFrame(top,sqldatabase, cfg, cond_on, extrabuttons)
    f.pack()
    top.wait_window(f)
    ret_val=top.commitclose         # =1 if committed (otherwise 0)
    top.destroy()
    return ret_val


""" 
editSQLTable_ntom_simple: link to tables via a simple n-to-m table (that contains only primary keys)
cfg must contain fields:
   ntom_table1, ntom_key1='student','key'                 : master table and its primary key
   ntom_table2,ntom_key2='minority','key'                 : slave table and key
   ntom_tablenm = 'minorityofstudent'                     : n-to-m table
   ntom_tablenm_cols=['key','key_student','key_minority'] : primary key, key name of table 1, key name of table 2 
"""
def editSQLTable_ntom_simple(parent, sqldatabase, title=None, cfg=None, key1_val=None):  # for parameters see SQLListboxForm_config
    assert cfg!=None and key1_val!=None, "editSQLTable_ntom_simple(parent,sqldatabase,titel,cfg,cond_on): Parameter cfg must not be 'None'"
    # (i) determine old selection
    tab1,key1                    = cfg.ntom_table1, cfg.ntom_key1
    tab2,key2                    = cfg.ntom_table2, cfg.ntom_key2
    tabnm, keynm, keynm1, keynm2 = cfg.ntom_tablenm, cfg.ntom_tablenm_cols[0], cfg.ntom_tablenm_cols[1], cfg.ntom_tablenm_cols[2]
    tabs    = [tab1,tab2,tabnm]    # list of all 3 tables
    join_on = [tab1+"."+key1+"="+tabnm+"."+keynm1, tab2+"."+key2+"="+tabnm+"."+keynm2]      # list of join conditions of the three tables
    wc=sqldatabase.getWhereClause_from_ColumnValues([tab1],[tab1+"."+key1],[key1_val])      # where-condition to select primary key key1_val 
    oldsel = sqldatabase.simple_select([tabnm+"."+keynm,tabnm+"."+keynm1,tabnm+"."+keynm2],tabs,join_on,wc) # old selection in format res=[(keynm,keyval1,keyval2,),(keynm,keyval1,keyval2,),...]   
    # (ii) invoke SQLListbox 
    oldsel_t2keys = [os[2] for os in oldsel]                                               # list of table2-keys (for preselection of listbox items)
    key2val_impossible=sqldatabase.get_new_primary_key(tab2,key2)[0]                       # impossible key for OR-Expression below (to avoid selection of all if oldsel_t2keys==[])
    where_init=sqldatabase.getWhereClause_from_ColumnValues([tab2],[tab2+"."+key2]+[tab2+"."+key2 for i in oldsel_t2keys],[key2val_impossible]+oldsel_t2keys,oper="OR")
    newsel = askSQLListboxForm(sqldatabase,cfg.link_ntom_listbox_cfg,where_init=where_init)[0] # new selection in format [(keyval2_0,val_0), (keyval2_1,val_1),...]
    if newsel==None:
        return 0
    else:
        newsel_t2keys = [ns[0] for ns in newsel]
        # (iii) determine items to be inserted and/or deleted 
        to_delete = []                   # keys of t2-items to be deleted
        for i in range(len(oldsel)):
            if not oldsel_t2keys[i] in newsel_t2keys:
                to_delete = to_delete + [oldsel_t2keys[i]]
        to_insert = []                   # t2-keys to be inserted (in association with key1=key1_val)
        for i in range(len(newsel)):
            if not newsel_t2keys[i] in oldsel_t2keys:
                to_insert = to_insert + [newsel_t2keys[i]]
        # (iv) insert/delete items (if necessary)
        for k in to_delete:
            print("deleting k=", k)
            sqldatabase.simple_delete(tabnm,[keynm1,keynm2],[key1_val,k])
        for k in to_insert:
            pk=sqldatabase.get_new_primary_key(tabnm)[0]
            sqldatabase.simple_insert(tabnm,[keynm,keynm1,keynm2],[pk,key1_val,k])
        return 1


##################################################################################################################################################################
##################################################################################################################################################################
# Part II: Module test
##################################################################################################################################################################
##################################################################################################################################################################
if __name__ == '__main__':
    # create a dummy database
    class tb_student_cfg(Table_cfg):
        name = 'student'
        # column definitions
        col_names           = ['id'         ,'name'        ,'address' ]
        col_types           = ['INTEGER'    ,'TEXT'        ,'TEXT'    ]
        col_references      = [None         ,None          ,None      ]
        col_constraints     = ['PRIMARY KEY','NOT NULL'    ,None      ]
        col_ref_constraints = [None         ,None          ,None      ]
        col_defaults        = [None         ,None          ,None      ]
        indexes             = [('idx_name',['name'])]
        # default rows (that will be stored during creation)
        rows_default     = ["(0,'N.N.'   , 'N.N.')", \
                            "(1,'Müller' , 'Musterweg 5    ; 72475 Musterstadt')", \
                            "(2,'Maier'  , 'Maierweg 6     ; 72473 Maierstadt' )", \
                            "(3,'Biegert', 'Biegertweg 6   ; 72471 Biegertstadt' )", \
                            "(4,'Schmitz', 'Schmitzweg 6   ; 72469 Schmitzstadt' )", \
                            "(5,'Alonso' , 'Schillerstr. 6 ; 72467 Knotendorf' )"]

    class tb_course_cfg(Table_cfg):
        name = 'course'
        # column definitions
        col_names           = ['id'         ,'course'     ]
        col_types           = ['INTEGER'    ,'TEXT'       ]
        col_references      = [None         ,None         ]
        col_constraints     = ['PRIMARY KEY','NOT NULL'   ]
        col_ref_constraints = [None         ,None         ]
        col_defaults        = [None         ,None         ]
        indexes             = [('idx_name',['course'])]
        # default rows (that will be stored during creation)
        rows_default     = ["(0,'N.N.'   )", \
                            "(1,'Mathe'  )", \
                            "(2,'English')", \
                            "(3,'Biology')"]

    class tb_exam_cfg(Table_cfg):
        name = 'exam'
        # column definitions
        col_names           = ['id'         ,'id_student'       ,'id_course','grade']
        col_types           = ['INTEGER'    ,'INTEGER'          ,'INTEGER'  ,'TEXT']
        col_references      = [None         ,'student'          ,'course'   ,None   ]
        #col_constraints     = ['PRIMARY KEY','NOT NULL'         ,'NOT NULL' ,None   ]
        col_constraints     = ['PRIMARY KEY',None               ,None       ,None   ]
        col_ref_constraints = [None         ,'ON DELETE CASCADE',None       ,None   ] 
        col_defaults        = [None         , 0                ,0       ,'"N.N."'        ]
        indexes             = None 
        # default rows (that will be stored during creation)
        rows_default     = ["(1, 2,1, 3.3)", \
                            "(2, 2,2, 4.0)", \
                            "(3, 4,3, 1.0)", \
                            "(4, 4,1, 2.0)", \
                            "(5, 3,1, 2.3)"]

    class db_dummy_cfg(sqldatabase_cfg):
        default_filename = 'sqldatabase_demo.db'
        table_configs = [tb_student_cfg, tb_course_cfg, tb_exam_cfg]

    print("\nModule test of supylib module supy.sqltableeditor.py")
    print("---------------------------------------------------------\n") 
    db = sqldatabase(db_dummy_cfg)
    db.print_database(1)

    
    class lb_EXAM_student_cfg(SQLListboxForm_config):
        tables = ['student']
        cols   = ['student.id', 'student.name']
        cols_format = ['5d','10s:10']
        cols_sortbutton = ['ID','NAME']
        where = None
        where_init = None
        colvals2str = colvals2str_default
        sep = ' | '
        width = 35
        height = 10
        lbfont = ('courier',12,'normal')
        button_str = ['Ok','Cancel']
        title      = 'lb_EXAM_student'

    class lb_EXAM_course_cfg(SQLListboxForm_config):
        tables = ['course']
        cols   = ['course.id', 'course.course']
        cols_format = ['5d','10s:10']
        cols_sortbutton = ['ID','NAME']
        where = None
        where_init = None
        colvals2str = colvals2str_default
        sep = ' | '
        width = 20
        height = 10
        lbfont = ('courier',12,'normal')
        button_str = ['Ok','Cancel']
        title      = 'lb_EXAM_course'

    class lb_EXAM_cfg(SQLListboxForm_config):
        tables = ['exam','student','course']
        join_on = ['student.id=exam.id_student', 'course.id=exam.id_course']
        cols   = ['exam.id', 'student.name', 'course.course']
        cols_format = ['5d','10s:10', '10s:10']
        cols_sortbutton = ['ID','STUD','COURSE']
        where = None
        where_init = None
        colvals2str = colvals2str_default
        sep = ' | '
        width = 40
        height = 10
        lbfont = ('courier',12,'normal')
        button_str = ['Ok','Cancel']
        title      = 'lb_EXAM'



    class db_sql_form_cfg(SupySQLForm_config):
        tables                = ['exam','student','course']
        join_on               = ['student.id=exam.id_student', 'course.id=exam.id_course']
        cols                  = ['exam.id', 'exam.id_student'  , 'exam.id_course', 'exam.grade']
        cols_type             = ['str'    , 'ref'              , 'ref'           , 'optionlist']
        cols_ref              = [None     , lb_EXAM_student_cfg, lb_EXAM_course_cfg, ['N.N.','1.0','1.3','1.7','2.0','2.3','2.7','3.0','3.3','3.7','4.0','4.7','5.0']]
        cols_readonly         = [1        ,0                , 0               , 0           ]
        cols_label            = ['ID','Student', 'Kurs', 'Note']
        cols_label_pos        = None
        cols_label_width      = 15     # may also be list
        cols_label_anchor     = 'w'    # may also be list
        cols_pos              = None
        cols_width            = 40     # may also be list
        cols_helptext         = ['Exam id...','Student name...', 'Kurs name...','Note ...']
        cols_browsebutton_pos = None
        select_cols           = ['exam.id']
        select_vals           = [2]
        help_event            = "<Button-3>"
        help_title            = "Help"
        help_buttonname       = "Close"
        browse_buttonname     = "Browse"
        checkResults = SupySQLForm_checkResults_default               # function to check if inputs are valid (see SupySQLForm_checkResults_default)
        update_tables = ['exam']                                      # tables to be updated after form submission
        update_tables_pkeys = [['exam.id']]                           # for each table (to be updated) a list of primary keys 
        update_tables_cols  = [['exam.id_student','exam.id_course','exam.grade']] # for each table (to be updated) a list of columns to be updated
        button_str = ['Ok','Cancel']                                  # names of the submit/cancel buttons
        title      = 'SupySQLForm...'                                 # is used only for askSupySQLForm

    db.print_database(1)
    res = askSupySQLForm(db,db_sql_form_cfg)
    print("Result:", res)
    db.print_database(1)

    # **************************
    class sql_table_editor_cfg(SQLTableEditor_config):
        table_cfg          = tb_exam_cfg        # table to be edited, first column is assumed to be primary key of type INTEGER!!!
        pkeys_readonly     = [0]                # records with primary key in this list are considered as READ_ONLY (e.g., for default records)
        cfg_choiceLB       = lb_EXAM_cfg        # configuration for choice listbox (typically a SQLListbox_config)
        cfg_recForm        = db_sql_form_cfg    # configuration for record form (typically a SupySQLForm_config)
        align_widgets      = 'horizontal'       # either 'vertical' or 'horizontal' for alignment of listbox and form
        lbButtons_str        = ['New', 'Copy', 'Delete']            # text for listbox buttons
        recFormButtons_str   = ['Reset', 'Default', 'Commit']       # text for form buttons
        ctrlRadioButtons_str = ['View', 'Edit']                     # text for radiobuttons (read-only, write/edit option)
        ctrlButtons_str      = ['Commit & Close', 'Cancel & Close'] # text for control buttons
        text_showinfo_cannot_delete_readonly = ("Info", "Cannot delete data record because it is read-only!")
        text_askyesno_delete = ("Delete Record?", "Do you really want to delete the selected record: ")
    #r=Tk()
    r=None 
    resxx = editSQLTables(r,db,"title hallo XXX",sql_table_editor_cfg)



#!/usr/bin/python
# -*- coding: utf-8 -*-

from tkinter import *
from supy.forms3 import *
from supy.utilities import *
from copy import *
import tkinter.messagebox
import tkinter.filedialog
from PIL import Image, ImageTk, ImageDraw


##################################################################################################################################################################
##################################################################################################################################################################
# Part I: SQLImageDialog and related classes and procedures:
#         classes: SQLImageDialog, SQLImageDialog_config, SQLImageDialogForm
#         SQLImageDialog: for inputing and clipping a image   
# Data Exchange:
##################################################################################################################################################################
##################################################################################################################################################################
class SQLImageDialog_config:
    text_labels    = ["Label","Filename","Mode","x1" ,"x2" ,"y1" ,"y2" ]
    pos_labels     = [(0,0)  ,(1,0)     ,(2,0)                 ,(3,0)  ,(4,0)  ,(5,0)  ,(6,0)  ]
    entry_width    = [40     ,40        ,None                  ,10     , 10    , 10    , 10    ]
    entry_pos      = [(0,1)  ,(1,1)     ,(2,1)                 ,(3,1)  ,(4,1)  ,(5,1)  ,(6,1)  ]   # format (row,column)
    entry_type     = [['ef'],['ef']     ,['ol',['full','clipped']],['ef'] ,['ef'] ,['ef'] ,['ef'] ]   # 'ef'=entry field; 'ol'=option list
    label_sticky   = 'w'                                                                           # Ausrichtung der Labels (within grid cell)
    entry_sticky   = 'w'                                                                           # Ausrichtung der Entries (within grid cell)
    title          = 'SQLImageDialog'
    button_str     = ['Browse', 'Browse', 'Ok','Apply','Cancel']                                   # Text on buttons 
    button_pos     = [[1,3],[3,3],None,None]
    image_canvas_size = [200, 100]                                                                 # Size of image canvas
    clipframe_mark_size = (5,2)           # side length and line width of marking rectangle                    
    clipframe_mark_color = ('red','red')  # color of square and rectangle of mark symbol

class SQLImageDialogFrame(Frame):
    """
    Default constructor
    """
    def __init__(self, parent, data_str_init=None, cfg=SQLImageDialog_config, paths_relative_to=os.getcwd()):
        # (i) do some basic initialization
        Frame.__init__(self,parent)  
        self.cfg=cfg                                        # configuration data
        self.paths_relative_to = paths_relative_to
        self.data_str=data_str_init                         # string representation of dialog result; format: "Label | Filename | mode | x1 | x2 | y1 | y2" (as could be saved in SQL table)
        self.data_list     = None
        self.data_label    = None                           # label to describe image
        self.data_filename = None                           # filename (inc. path) of image
        self.data_mode     = None                           # either 'full' or 'clipped': in former case the raw image is used, in the latter case a clip (see below) 
        self.data_x1       = None                           # (x1,x2,y1,y2) define a rectangle to clip from the original image
        self.data_x2       = None
        self.data_y1       = None
        self.data_y2       = None
        # (ii) allocate input widgets
        n_ip = len(self.cfg.text_labels)                     # number of input widgets
        self.leftframe = Frame(self, bd=2, relief=GROOVE)
        self.leftframe.pack(side=LEFT)
        self.rightframe = Frame(self)
        self.rightframe.pack(side=RIGHT,expand=YES, fill=BOTH)
        Button(self.leftframe,text=self.cfg.button_str[0],command=self.onBrowseFilename).grid(row=self.cfg.button_pos[0][0],column=self.cfg.button_pos[0][1])
        Button(self.leftframe,text=self.cfg.button_str[1],command=self.onBrowseClip    ).grid(row=self.cfg.button_pos[1][0],column=self.cfg.button_pos[1][1])        
        for i in range(n_ip): Label(self.leftframe,text=self.cfg.text_labels[i],anchor=self.cfg.label_sticky).grid(row=self.cfg.pos_labels[i][0],column=self.cfg.pos_labels[i][1],sticky=self.cfg.label_sticky)  # labels
        self.entry_vars = [StringVar() for i in range(n_ip)] # string-type entry variables 
        self.entry_field = [None for i in range(n_ip)]
        for i in range(n_ip):
            if   self.cfg.entry_type[i][0]=='ef':      # entry-field?
                entry = Entry(self.leftframe, textvariable=self.entry_vars[i], width=self.cfg.entry_width[i], bg="white", disabledforeground='black')
            elif self.cfg.entry_type[i][0]=='ol':      # option-list?
                entry = OptionMenu(self.leftframe,self.entry_vars[i],*self.cfg.entry_type[i][1])
            else:
                assert 0, "Wrong entry type: self.cfg.entry_type[i][0]="+str(self.cfg.entry_type[i][0])+" for i="+str(i)
            entry.grid(row=self.cfg.entry_pos[i][0],column=self.cfg.entry_pos[i][1],sticky=self.cfg.entry_sticky)
            self.entry_field[i]=entry
        self.dictEntryVar = {'label':self.entry_vars[0], 'fname':self.entry_vars[1], 'mode':self.entry_vars[2], \
                             'x1':self.entry_vars[3], 'x2':self.entry_vars[4], 'y1':self.entry_vars[5], 'y2':self.entry_vars[6]}
        # (iii) allocate image widgets (on rightframe)
        self.imcanvas = Canvas(self.rightframe,relief=SUNKEN)
        self.imcanvas.config(width=self.cfg.image_canvas_size[0], height=self.cfg.image_canvas_size[1])
        self.imcanvas.config(scrollregion=(0,0,self.cfg.image_canvas_size[0],self.cfg.image_canvas_size[1]))
        self.sbarx, self.sbary = Scrollbar(self.rightframe,orient=HORIZONTAL), Scrollbar(self.rightframe,orient=VERTICAL)
        self.sbarx.config(command=self.imcanvas.xview)
        self.sbary.config(command=self.imcanvas.yview)
        self.imcanvas.config(xscrollcommand=self.sbarx.set,yscrollcommand=self.sbary.set)
        self.sbarx.pack(side=BOTTOM, fill=X)
        self.sbary.pack(side=RIGHT, fill=Y)
        self.imcanvas.pack(side=RIGHT,expand=YES,fill=BOTH)
        self.imcanvas.bind("<Button-1>",self.onPressedB1)
        self.imcanvas.bind("<B1-Motion>",self.onMovedB1)
        self.imcanvas.bind("<ButtonRelease-1>",self.onReleasedB1)
        self.phim, self.draw = None, None
        self.imcanvas_state='IDLE'
        # (iv) set data 
        self.setData_from_str(self.data_str)           # set data fields from self.data_str
        self.sbarCenterView()                          # center view on clip (i.e. adjust srollbars accordingly)

    def setData_from_str(self,data_str):                         # split data_str in components and save in object fields
        if data_str:
            s=data_str.split('|')
            assert len(s)==7, "Assertion error in SQLImageDialog.setData_from_str(data_str): Parameter data_str='" + data_str + "' must have format 'Label | Filename | mode | x1 | x2 | y1 | y2' !"
            self.data_label   =str(s[0])
            self.data_filename=str(s[1])
            self.data_mode    =str(s[2])
            self.data_x1      =int(s[3])
            self.data_x2      =int(s[4])
            self.data_y1      =int(s[5])
            self.data_y2      =int(s[6])
            self.data_str = data_str
            self.data_list = [self.data_label, self.data_filename, self.data_mode, self.data_x1, self.data_x2, self.data_y1, self.data_y2]
        else:
            self.data_label, self.data_filename, self.data_mode = "", "",str(self.cfg.entry_type[2][0])
            self.data_x1, self.data_x2, self.data_y1, self.data_y2 = 0,0,0,0
        self.dictEntryVar['label'].set(self.data_label)
        self.dictEntryVar['fname'].set(self.data_filename)
        self.dictEntryVar['mode' ].set(self.data_mode)
        self.dictEntryVar['x1'   ].set(str(self.data_x1))
        self.dictEntryVar['x2'   ].set(str(self.data_x2))
        self.dictEntryVar['y1'   ].set(str(self.data_y1))
        self.dictEntryVar['y2'   ].set(str(self.data_y2))
        self.setImage()

    def setData_from_form(self):                               # evaluate form inputs and store them in result data fields
        self.data_label    = str(self.dictEntryVar['label'].get())
        self.data_filename = str(self.dictEntryVar['fname'].get())
        self.data_mode     = str(self.dictEntryVar['mode' ].get())
        self.data_x1       = int(self.dictEntryVar['x1'   ].get())
        self.data_x2       = int(self.dictEntryVar['x2'   ].get())
        self.data_y1       = int(self.dictEntryVar['y1'   ].get())
        self.data_y2       = int(self.dictEntryVar['y2'   ].get())
        self.data_str      = self.data_label + '|' + self.data_filename + '|' + self.data_mode + '|' + str(self.data_x1) + '|' + str(self.data_x2) + '|' + str(self.data_y1) + '|' + str(self.data_y2) 
        self.data_list = [self.data_label, self.data_filename, self.data_mode, self.data_x1, self.data_x2, self.data_y1, self.data_y2]

    def load_image(self,fname):   # load image with filename fname and set to self.phim
        self.phim,self.draw=None,None
        if(fname):
            try:
                self.image = Image.open(search_path(fname,self.paths_relative_to))
            except IOError as e:
                msg="I/O error({0}): {1}".format(e.errno, e.strerror) + "\nsupy.sqlforms.SQLImageDialogFrame.load_image: Cannot load image " + fname + " !"
                tkinter.messagebox.showerror(title="Submit Error", parent=self, message=msg)
            else:
                self.draw = ImageDraw.Draw(self.image)
                self.phim = ImageTk.PhotoImage(self.image)
        return self.phim,self.draw

    def setImage(self):
        # (i) load image
        self.load_image(self.data_filename)
        # (ii) draw image on canvas
        if(self.imcanvas!=None):
            self.imcanvas.delete('all');
            szx,szy=0,0
            if(self.phim!=None):
                szx,szy=self.phim.width(), self.phim.height()
                self.imcanvas.config(scrollregion=(0,0,szx,szy))
                self.imcanvas.create_image(0,0,image=self.phim,anchor=NW)
                if(self.dictEntryVar['mode'].get()=='clipped'):
                    self.imcanvas.create_rectangle(self.data_x1,self.data_y1,self.data_x2,self.data_y2,width=self.cfg.clipframe_mark_size[1],outline=self.cfg.clipframe_mark_color[1])
                    dxy=int(self.cfg.clipframe_mark_size[0]*0.5) 
                    self.imcanvas.create_rectangle(self.data_x1-dxy,self.data_y1-dxy,self.data_x1+dxy,self.data_y1+dxy,width=self.cfg.clipframe_mark_size[1],outline=self.cfg.clipframe_mark_color[0])
                    self.imcanvas.create_rectangle(self.data_x2-dxy,self.data_y2-dxy,self.data_x2+dxy,self.data_y2+dxy,width=self.cfg.clipframe_mark_size[1],outline=self.cfg.clipframe_mark_color[0])

    def sbarCenterView(self):     # center scrollbar position on clipped image (if necessary)
        x1,x2,y1,y2 = self.data_x1, self.data_x2, self.data_y1, self.data_y2     # bounding box clipped image
        dx,dy       = abs(x2-x1), abs(y2-y1)                                     # width, height of clipped image
        imx,imy     = self.phim.width(), self.phim.height()                      # width, height of original image
        cvx,cvy     = self.imcanvas.winfo_width(), self.imcanvas.winfo_height()  # canvas size
        if(cvx<2) : cvx=self.imcanvas.winfo_reqwidth()
        if(cvy<2) : cvy=self.imcanvas.winfo_reqheight()
        fx,fy,padx,pady=0,0,0,0   
        if(x2>cvx)and(imx>cvx):
            padx=int(0.5*(cvx-dx))
            if(padx<0):
                padx=0
            fx=(x1-padx)/imx
        if(y2>cvy)and(imy>cvy):
            pady=int(0.5*(cvy-dy))
            if(pady<0):
                pady=0
            fy=(y1-pady)/imy
        self.imcanvas.xview_moveto(fx)
        self.imcanvas.yview_moveto(fy)

    def onBrowseFilename(self):
        s=self.dictEntryVar['fname'].get()   
        filename  = os.path.basename(s)
        directory = os.path.dirname(s)
        if directory=="":
            directory="."
        s=tkinter.filedialog.askopenfilename(initialdir=directory, initialfile=filename)
        if(s!=None)and(s!="")and(s):
            s=os.path.relpath(s,self.paths_relative_to)
            self.dictEntryVar['fname'].set(s)
            self.data_filename=s
            self.setImage()

    def onPressedB1(self,event):    # pressed left button on image canvas
        # check if clicked on lower-right (x2,y2)-mark
        szx,szy=self.phim.width(), self.phim.height()
        x,y=int(event.x+szx*self.sbarx.get()[0]),int(event.y+szy*self.sbary.get()[0])
        tol=2*self.cfg.clipframe_mark_size[0]
        if (abs(x-self.data_x1)<tol) and (abs(y-self.data_y1)<tol):
            # click on upper-left mark: then move upper-left mark
            self.imcanvas_state='MOVE_UL'
            self.onMovedB1(event)
        elif (abs(x-self.data_x2)<tol) and (abs(y-self.data_y2)<tol):
            # click on lower-right mark: then move lower-right mark
            self.imcanvas_state='MOVE_LR'
            self.onMovedB1(event)
        else:
            # else: set upper-left and then move lower-right mark
            self.dictEntryVar['x1'].set(str(x))
            self.dictEntryVar['y1'].set(str(y))
            self.data_x1,self.data_y1=x,y
            self.imcanvas_state='MOVE_LR'
            self.onMovedB1(event)

    def onMovedB1(self,event):      # moved left button on image canvas
        if(self.imcanvas_state!='IDLE'):
            szx,szy=self.phim.width(), self.phim.height()
            x,y=int(event.x+szx*self.sbarx.get()[0]),int(event.y+szy*self.sbary.get()[0])
            if(self.imcanvas_state=='MOVE_LR'):
                # move lower-right mark
                self.dictEntryVar['x2'].set(str(x))
                self.dictEntryVar['y2'].set(str(y))
                self.data_x2,self.data_y2=x,y
            else: 
                # move upper-left mark
                self.dictEntryVar['x1'].set(str(x))
                self.dictEntryVar['y1'].set(str(y))
                self.data_x1,self.data_y1=x,y
            self.setImage()

    def onReleasedB1(self,event):   # released left button on image canvas
        self.onMovedB1(event)
        self.imcanvas_state='IDLE'

    def onBrowseClip(self):
        pass

    def onApply(self):
        self.setData_from_form()
        self.setImage()


class SQLImageDialog:
    def __init__(self, parent, data_str_init=None, cfg=SQLImageDialog_config):           # for parameters see SQLListboxForm_config
        # basic initialization
        self.cfg = cfg
        self.data_list=None
        self.data_str =None
        self.box = Frame(parent)                             # container frame
        self.box.pack(expand=YES, fill=X)
        # create ImageDialogFrame
        self.imdia = SQLImageDialogFrame(self.box,data_str_init,cfg) 
        self.imdia.pack(side=TOP) #, expand=Y, fill=X)      
        # create control button frame
        self.buttonframe = Frame(self.box)
        Button(self.buttonframe,text=self.cfg.button_str[4],command=self.onCancel).pack(side=RIGHT)
        Button(self.buttonframe,text=self.cfg.button_str[2],command=self.onSubmit).pack(side=RIGHT)
        Button(self.buttonframe,text=self.cfg.button_str[3],command=self.onApply).pack(side=LEFT)
        self.box.master.bind('<Return>', (lambda event: self.onSubmit()))
        self.buttonframe.pack(side=TOP)

    def onSubmit(self):                                      # override this if necessary
        self.imdia.setData_from_form()
        self.data_list = self.imdia.data_list 
        self.data_str  = self.imdia.data_str 
        self.box.destroy()
     
    def onApply(self):
        self.imdia.onApply()

    def onCancel(self):                                      # override if needed
        self.data_list=None
        self.data_str=None
        self.box.destroy()
 
def askSQLImageDialog(cfg=SQLImageDialog_config, parent=None, title=None, data_str_init=None):
    if title==None: title=cfg.title
    top=Toplevel(parent)
    top.grab_set()                         # make window modal (i.e., redirect all inputs from parent window to listbox dialog)
    if parent: top.transient(parent)       # make window transient, e.g., minimize with parent etc.
    if title: top.title(title) 
    f=SQLImageDialog(top,data_str_init,cfg)
    top.wait_window(f.box)
    top.destroy()
    return f.data_str, f.data_list   # return results
    

##################################################################################################################################################################
##################################################################################################################################################################
# Part II: ScrolledText: A Tkinter Text Widget with scrollbars
##################################################################################################################################################################
##################################################################################################################################################################

class ScrolledTextFrame(Frame):
    def __init__(self,parent=None,text_str=''):
        Frame.__init__(self,parent)
        self.text=Text(self,relief=SUNKEN)
        self.sbarx, self.sbary = Scrollbar(self,orient=HORIZONTAL), Scrollbar(self,orient=VERTICAL)
        self.sbarx.config(command=self.text.xview)
        self.sbary.config(command=self.text.yview)
        self.text.config(xscrollcommand=self.sbarx.set,yscrollcommand=self.sbary.set)
        self.sbarx.pack(side=BOTTOM, fill=X)
        self.sbary.pack(side=RIGHT, fill=Y)
        self.settext(text_str)

    def settext(self,text_str=''):        # set text given by text_str
        self.text.delete('1.0',END)
        self.text.insert('1.0',text_str)
        self.text.mark_set(INSERT, END)
        self.text.focus()

    def gettext(self):                    # return text of text-widget
        return self.text.get('1.0', END+'-1c')    # first through last
        


##################################################################################################################################################################
##################################################################################################################################################################
# Part IV: Module test
##################################################################################################################################################################
##################################################################################################################################################################
if __name__ == '__main__':
    # create a dummy database
    class tb_student_cfg(Table_cfg):
        name = 'student'
        # column definitions
        col_names           = ['id'         ,'name'        ,'address' ]
        col_types           = ['INTEGER'    ,'TEXT'        ,'TEXT'    ]
        col_references      = [None         ,None          ,None      ]
        col_constraints     = ['PRIMARY KEY','NOT NULL'    ,None      ]
        col_ref_constraints = [None         ,None          ,None      ]
        col_defaults        = [None         ,None          ,None      ]
        indexes             = [('idx_name',['name'])]
        # default rows (that will be stored during creation)
        rows_default     = ["(1,'Müller' , 'Musterweg 5    ; 72475 Musterstadt')", \
                            "(2,'Maier'  , 'Maierweg 6     ; 72473 Maierstadt' )", \
                            "(3,'Biegert', 'Biegertweg 6   ; 72471 Biegertstadt' )", \
                            "(4,'Schmitz', 'Schmitzweg 6   ; 72469 Schmitzstadt' )", \
                            "(5,'Alonso' , 'Schillerstr. 6 ; 72467 Knotendorf' )"]

    class tb_course_cfg(Table_cfg):
        name = 'course'
        # column definitions
        col_names           = ['id'         ,'course'     ]
        col_types           = ['INTEGER'    ,'TEXT'       ]
        col_references      = [None         ,None         ]
        col_constraints     = ['PRIMARY KEY','NOT NULL'   ]
        col_ref_constraints = [None         ,None         ]
        col_defaults        = [None         ,None         ]
        indexes             = [('idx_name',['course'])]
        # default rows (that will be stored during creation)
        rows_default     = ["(1,'Mathe'  )", \
                            "(2,'English')", \
                            "(3,'Biology')"]

    class tb_exam_cfg(Table_cfg):
        name = 'exam'
        # column definitions
        col_names           = ['id'         ,'id_student'       ,'id_course','grade']
        col_types           = ['INTEGER'    ,'INTEGER'          ,'INTEGER'  ,'FLOAT']
        col_references      = [None         ,'student'          ,'course'   ,None   ]
        col_constraints     = ['PRIMARY KEY','NOT NULL'         ,'NOT NULL' ,None   ]
        col_ref_constraints = [None         ,'ON DELETE CASCADE',None       ,None   ] 
        col_defaults        = [None         ,None               ,None       ,None    ]
        indexes             = None 
        # default rows (that will be stored during creation)
        rows_default     = ["(1, 2,1, 3.3)", \
                            "(2, 2,2, 4.0)", \
                            "(3, 4,3, 1.0)", \
                            "(4, 4,1, 2.0)", \
                            "(5, 3,1, 2.3)"]

    class db_dummy_cfg(sqldatabase_cfg):
        default_filename = 'sqldatabase_demo.db'
        table_configs = [tb_student_cfg, tb_course_cfg, tb_exam_cfg]
        
    print("\nModule test of supylib module supy.sqlforms.py")
    print("--------------------------------------------------\n") 
    db = sqldatabase(db_dummy_cfg)
    db.print_database(1)
    root = Tk()

    ##################################################################################################################################################################
    print("Part I: Test of SQLImageDialog:")
    print("--------------------------------")
    ##################################################################################################################################################################
    class mySQLImageDialog_config(SQLImageDialog_config):
        text_labels    = ["Label","Filename","Mode","x1" ,"x2" ,"y1" ,"y2" ]
        pos_lables     = [(0,0)  ,(1,0)     ,(2,0)                 ,(3,0)  ,(4,0)  ,(5,0)  ,(6,0)  ]
        entry_width    = [40     ,40        ,None                  ,10     , 10    , 10    , 10    ]
        entry_pos      = [(0,1)  ,(1,1)     ,(2,1)                 ,(3,1)  ,(4,1)  ,(5,1)  ,(6,1)  ]   # format (row,column)
        entry_type     = [['ef'],['ef']     ,['ol',['full','clipped']],['ef'] ,['ef'] ,['ef'] ,['ef'] ]   # 'ef'=entry field; 'ol'=option list
        label_sticky   = 'w'                                                                           # Ausrichtung der Labels (within grid cell)
        entry_sticky   = 'w'                                                                           # Ausrichtung der Entries (within grid cell)
        title          = 'SQLImageDialog'
        button_str     = ['Browse', 'Browse', 'Ok','Apply','Cancel']                                   # Text on buttons 
        button_pos     = [[1,3],[3,3],None,None]
        image_canvas_size = [500, 300]                                                                 # Size of image canvas
        clipframe_mark_size = (5,2)           # side length and line width of marking rectangle                    
        clipframe_mark_color = ('red','red')  # color of square and rectangle of mark symbol

    res_str, res_list = askSQLImageDialog(mySQLImageDialog_config,parent=root,title="Mein Titel...",data_str_init="Studenten WS 13/14|bild1.jpg|clipped|1180|1300|820|846")
    print("Result: res_str=", res_str, " res_list=", res_list)

    ##################################################################################################################################################################
    print("Part II: Test of SQLListboxForm:")
    print("--------------------------------")
    ##################################################################################################################################################################

    class db_sqllb_form_cfg(SQLListboxForm_config):
        tables = None
        join_on = None
        cols   = None
        cols_format = None
        cols_sortbutton = None
        cols_sortbutton = None
        where = None
        where_init = None
        colvals2str = colvals2str_default
        sep = ' | '
        width = 20
        height = 10
        lbfont = ('courier',12,'normal')
        mode_multiselect = True
        callback_select = callback_select_default
        button_str = ['Ok','Cancel']
        title      = 'SQLListboxForm...'

    #lb = SQLListbox(root,db,['exam','student'],['exam.id','student.name','student.address'],\
    #                cols_format=['3d','3s:5','10s:10'],cols_sortbutton=['exxID','stName','stAdr'],labeltext='SQLListbox',\
    #                width=40, height=15)
    #lb.pack()
    #root.mainloop()
    res,res_str = askSQLListboxForm(db,db_sqllb_form_cfg, \
                                    ['exam','student'],['exam.id_student==student.id'],['exam.id','student.name','student.address'],\
                                    root,"SQLListboxForm-Dialog...",\
                                    cols_format=['3d','3s:5','10s:10'],cols_sortbutton=['exxID','stName','stAdr'],labeltext='SQLListbox',\
                                    width=40, height=15)
    print("Result:", res, res_str)

    ##################################################################################################################################################################
    print("Part IIIa: Test of SupySQLForm:")
    print("--------------------------------")
    ##################################################################################################################################################################

    class db_sql_form_cfg(SupySQLForm_config):
        tables                = ['student']
        join_on               = None
        cols                  = ['student.id','student.name']
        cols_type             = ['str','filename']
        cols_ref              = None
        cols_readonly         = [0,0]
        cols_label            = ['ID','Name']
        cols_label_pos        = None
        cols_label_width      = 15     # may also be list
        cols_label_anchor     = 'w'    # may also be list
        cols_pos              = None
        cols_width            = 40     # may also be list
        cols_helptext         = ['Student id...','Student name...']
        cols_browsebutton_pos = None
        select_cols           = ['student.name']
        select_vals           = ['Maier']
        help_event            = "<Button-3>"
        help_title            = "Help"
        help_buttonname       = "Close"
        browse_buttonname     = "Browse"
        checkResults = SupySQLForm_checkResults_default               # function to check if inputs are valid (see SupySQLForm_checkResults_default)
        update_tables = ['student']                                   # tables to be updated after form submission
        update_tables_pkeys = [['student.id']]                        # for each table (to be updated) a list of primary keys 
        update_tables_cols  = [['student.name']]                      # for each table (to be updated) a list of columns to be updated
        button_str = ['Ok','Cancel']                                  # names of the submit/cancel buttons
        title      = 'SupySQLForm...'                                 # is used only for askSupySQLForm
    res = askSupySQLForm(db,db_sql_form_cfg)
    print("Result:", res)

    ##################################################################################################################################################################
    print("Part IIIb: Advanced Test of SupySQLForm:")
    print("-----------------------------------------")
    ##################################################################################################################################################################

    class lb_EXAM_student_cfg(SQLListboxForm_config):
        tables = ['student']
        cols   = ['student.id', 'student.name']
        cols_format = ['5d','10s:10']
        cols_sortbutton = ['ID','NAME']
        where = None
        where_init = None
        colvals2str = colvals2str_default
        sep = ' | '
        width = 40
        height = 10
        lbfont = ('courier',12,'normal')
        callback_select = callback_select_default
        button_str = ['Ok','Cancel']
        title      = 'lb_EXAM_student'

    class lb_EXAM_course_cfg(SQLListboxForm_config):
        tables = ['course']
        cols   = ['course.id', 'course.course']
        cols_format = ['5d','10s:10']
        cols_sortbutton = ['ID','NAME']
        where = None
        where_init = None
        colvals2str = colvals2str_default
        sep = ' | '
        width = 20
        height = 10
        lbfont = ('courier',12,'normal')
        callback_select = callback_select_default
        button_str = ['Ok','Cancel']
        title      = 'lb_EXAM_course'

    class db_sql_form_cfg(SupySQLForm_config):
        tables                = ['exam','student','course']
        cols                  = ['exam.id', 'exam.id_student'  , 'exam.id_course', 'exam.grade']
        cols_type             = ['str'    , 'ref'              , 'ref'           , 'optionlist']
        cols_ref              = [None     , lb_EXAM_student_cfg, lb_EXAM_course_cfg, ['1.2','2.3','4.0','5.0']]
        cols_readonly         = [1        ,0                , 0               , 0           ]
        cols_label            = ['ID','Student', 'Kurs', 'Note']
        cols_label_pos        = None
        cols_label_width      = 15     # may also be list
        cols_label_anchor     = 'w'    # may also be list
        cols_pos              = None
        cols_width            = 40     # may also be list
        cols_helptext         = ['Exam id...','Student name...', 'Kurs name...','Note ...']
        cols_browsebutton_pos = None
        select_cols           = ['exam.id']
        select_vals           = [2]
        help_event            = "<Button-3>"
        help_title            = "Help"
        help_buttonname       = "Close"
        browse_buttonname     = "Browse"
        checkResults = SupySQLForm_checkResults_default               # function to check if inputs are valid (see SupySQLForm_checkResults_default)
        update_tables = ['exam']                                      # tables to be updated after form submission
        update_tables_pkeys = [['exam.id']]                           # for each table (to be updated) a list of primary keys 
        update_tables_cols  = [['exam.id_student','exam.id_course','exam.grade']] # for each table (to be updated) a list of columns to be updated
        button_str = ['Ok','Cancel']                                  # names of the submit/cancel buttons
        title      = 'SupySQLForm...'                                 # is used only for askSupySQForm

    db.print_database(1)
    res = askSupySQLForm(db,db_sql_form_cfg)
    print("Result:", res)
    db.print_database(1)




#!/usr/bin/python
# -*- coding: utf-8 -*-

############################################################
# a reusable form class
############################################################
     
from Tkinter import *
import tkMessageBox
import tkFileDialog
import os,time,datetime

class SupyForm:                                           # add non-modal form box
    def __init__(self, labels, default_entries=None, parent=None,labelanchor='w',labelwidth=None,entrysize=40,pos=None,
                 flagButtons=1,options=None,paths_relative_to=os.getcwd(),browse_callback=None,
                 help_event="<F1>",help_texts=None,help_callback=None):        # pass field labels list
        '''
        labels: list of strings where labels[i] is the label of the i-th input field
        default_entries: list of default entries that are displayed initially in the input field
        parent: parent widget
        labelanchor: aliggnment of label
        labelwidth: width of label fields, either a None, a single number, or a list (for each input field)
        entrysize: width of entry fields, either a single number or a list (for each input field)
        pos: either None or list; pos[i]=(x,y) is grid position for i-th label etc.  
        flagButtons: If >0 then buttons are displayed for submit/cancel actions 
        options: either None or a list where options[i] is either 'filename' or None or a string list for a Optionmenu
        paths_relative_to: save browsed paths relative to this path
        browse_callback(idx): function that is called after selecting a filename, parameter idx is field index
        help_event: event that evokes help function (e.g., pressing F1 over the label)
        help_texts: either None or list of help strings for each label
        help_call : either None (then default call is used) or function 
        '''
        self.parent=parent
        if(pos==None):
            pos = [(0,y) for y in range(len(labels))]
        if(not labelwidth):
            labelwidth=max([len(s) for s in labels])
        if not isinstance(labelwidth,list):
            labelwidth = [labelwidth for y in range(len(labels))]
        if not isinstance(entrysize,list):
            entrysize = [entrysize for y in range(len(labels))]
        self.options = options
        self.paths_relative_to = paths_relative_to
        self.browse_callback=browse_callback
        self.browsebuttons = []
        self.labels = labels
        box = Frame(parent)
        box.pack(expand=YES, fill=X)
        formbox = Frame(box, bd=2, relief=GROOVE)     # formbox will have grid layout
        self.formbox=formbox
        formbox.pack(side=TOP, expand=Y, fill=X)      # runs onSubmit method
        self.content = {}                 # dictionary of input entries
        self.results = None               # final results as set by onSubmit()
        self.entry_vars = [StringVar() for l in labels]
        for sv in self.entry_vars: sv.set("")
        for i in range(len(labels)):
            row_label,column_label=pos[i][1], pos[i][0]
            lb=Label(formbox, text=labels[i], anchor=labelanchor, width=labelwidth[i])
            lb.grid(row=row_label,column=column_label)
            if(help_texts!=None) and (help_texts[i]!=None):
                if(help_callback==None):
                    help_callback=self.displayHelpMessage_default
                lb.bind(help_event,lambda ev,txt=help_texts[i]: help_callback(txt))
            if(options==None)or(options[i]==None)or(options[i]=='filename')or(options[i]==[]):                
                entry = Entry(formbox, textvariable=self.entry_vars[i], width=entrysize[i], 
                              bg="white", disabledforeground='black')
            else:
                entry = OptionMenu(formbox,self.entry_vars[i],*options[i])
                entry.config(width=int(0.88*entrysize[i]), disabledforeground='black')
            self.content[labels[i]] = entry
            if default_entries and (len(default_entries)>=(i+1)):
                self.setEntryText(i,default_entries[i])                
            entry.grid(row=row_label,column=column_label+1) # side=TOP, expand=YES, fill=X)
            if(options!=None)and(options[i]=='filename'):
                b=Button(formbox,text='Browse',command=(lambda i=i,label=labels[i]:self.onBrowse(i,label)))
                b.grid(row=row_label,column=column_label+2)
                self.browsebuttons.append(b)   
        if(flagButtons>0):
            Button(box, text='Cancel', command=self.onCancel).pack(side=RIGHT)
            Button(box, text='Submit', command=self.onSubmit).pack(side=RIGHT)
            box.master.bind('<Return>', (lambda event: self.onSubmit()))
        self.box=box
        self.formbox=formbox
        self.labels=labels

    def displayHelpMessage_default(self,text):
        w=Toplevel(self.parent)
        if self.parent: w.transient(self.parent)     # make window transient, e.g., minmize with parent etc.
        w.title("Help")
        l=Label(w,text=text)
        l.pack()
        b=Button(w,text="Close",command=w.destroy)
        b.pack()
        #tkMessageBox.showinfo("Help", text, parent=w)

    def setEntryText(self,idx,txt):
        if isinstance(self.content[self.labels[idx]],Entry):
            self.entry_vars[idx].set(txt)
            self.content[self.labels[idx]].configure(textvariable=self.entry_vars[idx])
        else:
            if (txt in self.options[idx]):
                self.entry_vars[idx].set(txt)

    def disableWidgets(self):
        for w in self.content.values():
            w.configure(state=DISABLED)
        for b in self.browsebuttons:
            b.configure(state=DISABLED)
        
    def enableWidgets(self):
        for w in self.content.values():
            w.configure(state=NORMAL)
        for b in self.browsebuttons:
            b.configure(state=NORMAL)
        
    def onBrowse(self,i,label):
        s=self.content[label].get()   
        filename  = os.path.basename(s)
        directory = os.path.dirname(s)
        if directory=="":
            directory="."
        s=tkFileDialog.askopenfilename(initialdir=directory, initialfile=filename)
        if(s!=None)and(s!="")and(s):
            s=os.path.relpath(s,self.paths_relative_to)
            self.setEntryText(i,s)
            if(self.browse_callback!=None):
                self.browse_callback(i)

    def onSubmit(self):                                      # override this
        for key in self.content.keys():                      # user inputs in 
            print key, '\t=>\t', self.content[key].get()     # self.content[k]
        self.results = [self.content[l].get() for l in self.labels] 
        self.box.destroy()
     
    def onCancel(self):                                      # override if need
        self.results=None
        self.box.destroy()
     
#class DynamicPyFeForm(PyFeForm):
#    def __init__(self, labels=None):
#        labels = raw_input('Enter field names: ').split()
#        Form.__init__(self, labels)
#    def onSubmit(self):
#        print 'Field values...'
#        Form.onSubmit(self)           
#        self.onCancel()
        
def askSupyForm(labels,default_entries=None,parent=None,title=None,labelanchor='w',labelwidth=None,entrysize=40):      
    top=Toplevel(parent)
    if parent: top.transient(parent)       # make window transient, e.g., minmize with parent etc.
    if title: top.title(title) 
    f=SupyForm(labels,default_entries,top,labelanchor,labelwidth,entrysize)
    top.wait_window(f.box)
    top.destroy()
    return f.results

# *****************************
# slider for dates
# *****************************
class DateSlider(Frame):
    def __init__(self, parent, from_date=datetime.date(2012,1,1),to_date=datetime.date(2012,12,31),init_date=datetime.date(2012,1,1),orient_arg='horizontal',command=None,*args):
        Frame.__init__(self,parent)
        dd=to_date-from_date
        self.var=IntVar()
        self.var.set((init_date-from_date).days)
        self.from_date=from_date
        self.current_date=init_date
        self.to_date=to_date
        self.command=command
        self.slider = Scale(self,from_=0,to_=dd.days,command=self.onMove,variable=self.var,*args,orient=orient_arg)
        self.slider.pack(side=RIGHT)
        self.label = Label(self,text="Halleluja")
        self.label.pack(side=LEFT)
        self.onMove(self.var.get(),1)
        
    def onMove(self,value,initFlag=0):
        self.current_date=self.from_date+datetime.timedelta(self.var.get())
        self.str_date = "%02d/%02d/%04d" % (self.current_date.month, self.current_date.day, self.current_date.year)
        self.label.configure(text=self.str_date)
        if (self.command !=None) and (initFlag==0):
            self.command(value)

    def get(self):
        return self.current_date


    
# *****************************
# slider for dates
# *****************************
class SupySliderForm:
    """
    Input form using Sliders
    Parameters:
       parent = master widget
       labels = list of name lables for sliders
       default_values = initial slider values (either float or date)
       ranges = list of triples (v_min,v_max,delta_v)
    """
    def __init__(self, parent, labels, default_values, ranges=None, labelanchor='w',command=None,labelwidth=None,sliderwidth=40):        # pass field labels list
        if(not labelwidth):
            labelwidth=max([len(s) for s in labels])
        box = Frame(parent)
        box.pack(expand=YES, fill=X)
        sliderbox=Frame(box)
        sliderbox.pack(side=TOP,expand=YES,fill=X)

        #rows = Frame(box, bd=2, relief=GROOVE)        # box has rows, button
        #lcol = Frame(rows)                            # rows has lcol, rcol
        #rcol = Frame(rows)                            # button or return key,
        #rows.pack(side=TOP, expand=Y, fill=X)         # runs onSubmit method
        #lcol.pack(side=LEFT)
        #rcol.pack(side=RIGHT, expand=Y, fill=X)
        self.labels=labels
        self.content = {}                 # dictionary of sliders
        self.results = None               # final results as set be onSubmit()
        for i in range(len(labels)):
            label = labels[i]
            Label(sliderbox, text=label+": ", anchor=labelanchor, width=labelwidth).grid(row=i,column=0)
            default_value=default_values[i]
            range_=ranges[i]
            if type(default_value)==datetime.date:
                slider=DateSlider(sliderbox,range_[0],range_[1],default_value,command=self.onUpdate)
            else:
                slider=Scale(sliderbox,from_=range_[0],to_=range_[1],resolution=range_[2],command=self.onUpdate,orient='horizontal')
                slider.set(default_value)
            slider.grid(row=i,column=1) #side=TOP, expand=YES, fill=X)
            self.content[label] = slider
        Button(box, text='Cancel', command=self.onCancel).pack(side=RIGHT)
        Button(box, text='Submit', command=self.onSubmit).pack(side=RIGHT)
        box.master.bind('<Return>', (lambda event: self.onSubmit()))
        self.box=box
        self.sliderbox=sliderbox
        self.command=command
   
    def onUpdate(self,value=0):                                   # override this
        self.results = [self.content[l].get() for l in self.labels] 
        if(self.command):
            self.command()
     
    def onSubmit(self):
        for key in self.content.keys():                      # user inputs in 
            print key, '\t=>\t', self.content[key].get()     # self.content[k]
        self.onUpdate()
        self.box.destroy()
     
    def onCancel(self):                                       # override if need
        self.results=None
        self.box.destroy()
    
def askSupySliderForm(labels,default_values,ranges,parent=None,title=None,labelanchor='w',command=None,labelwidth=None,sliderwidth=40):      
    top=Toplevel(parent)
    if parent: top.transient(parent)       # make window transient, e.g., minmize with parent etc.
    if title: top.title(title) 
    f=SupySliderForm(top,labels,default_values,ranges,labelanchor,command,labelwidth,sliderwidth)
    top.wait_window(f.box)
    top.destroy()
    return f.results

if __name__ == '__main__':
    root=Tk()
    s=DateSlider(root)
    s.pack(side=TOP)
    # Example 1: SupyFrom
    labels=['Name', 'Age', 'Job']
    defaults=['Fridolin','40',u"Großer Mandarin"]
    r=askSupyForm(labels,defaults,root,"Python ask form ...",'e',30,40)
    print r
    # Example 2: SupySliderFrom
    labels=['var1','var2','var3','date']
    defaults=[3,14,5,datetime.date(2012,11,21)]
    ranges=[(0,10,1),(10,20,2),(5,10,0.1),(datetime.date(2012,1,1),datetime.date(2012,12,31))]
    r=askSupySliderForm(labels,defaults,ranges,root,"Slider Form example",'e',None,30,40)
